from .compress_util import glue_seq

def lz20(data):
    """Lempel-Ziv 2020 :D using the naming convention of lz77 and lz78.
       The idea of the algorithm that it looks for sequences in the data as in lz78.
       But instead of creates the dictionary and assigning codes while compressing,
       it glues the sequences together for the huffman coding to code them based on occurrences."""

    data = list(map(str, data))
    data = [i.replace(" ", "__") for i in data]
    i = 1
    sequences = []
    start = 0
    n_data = len(data)
    seq_dict = {}

    while i <= n_data:
        s = glue_seq(data[start:i])
        ls = len(s.split())
        if s in seq_dict:
            if ls + start >= n_data:
                # if it is the final seq just append it
                sequences.append(s)
                seq_dict[s] += 1
                break  # ensure that it doesn't loop forever
            seq_dict[s] += 1
            i += 1
            continue
        else:
            seq_dict[s] = 1
            if " " in s:  # it is a sequence of symbols not one symbol
                h, t = glue_seq(s, last_separate = True)
                sequences.append(h)
                sequences.append(t)
            else:
                sequences.append(s)
            start += ls
            i += 1

    return sequences



from .compress_util import glue_seq

def lz78(data):
    """Normal Lempel-Ziv78 which assigns codes for each new encountered sequence."""

    data = list(map(str, data))
    i = 1
    sequences = []
    start = 0
    n_data = len(data)
    seq_dict = {}
    code = 1

    while i <= n_data:
        s = glue_seq(data[start:i])
        ls = len(s.split())
        if s in seq_dict:
            if ls + start >= n_data:
                sequences.append(s)
                break
            i += 1
            continue
        else:
            seq_dict[s] = str(code)
            if " " in s:
                h, t = glue_seq(s, last_separate = True)
                sequences.append(seq_dict[h])
                sequences.append(t)
            else:
                sequences.append(seq_dict[s])
            start += ls
            i += 1
            code += 1

    return sequences, seq_dict



from math import floor, ceil
from cpython cimport int as Integer

ctypedef fused str_or_byte:
    bytes
    str

cdef str bit_fill(list comp):
    cdef Integer i
    cdef str s = "1" + "".join([bin(i)[2:].zfill(9) for i in comp])
    return s

cdef Integer byte_encode(str bits):
    cdef Integer i = int(bits, 2)
    return i

cdef str byte_decode(str_or_byte enc):
    cdef str bins = bin(int.from_bytes(enc, "big"))[2:].zfill(len(enc) * 8)
    return bins

cdef list byte2int(str bits, int ext_bytes):
    cdef int i
    cdef list lst = [int(bits[i * 9:(i + 1) * 9], 2) for i in range(ext_bytes)]
    return lst

cdef int i, ib
cdef dict ascii_to_int = {i.to_bytes(1, "big"): i for i in range(256)}

cdef bytes b
cdef dict int_to_ascii = {ib: b for b, ib in ascii_to_int.items()}

## Lempel-Ziv-Welch algorithm to compress the symbol sequences in the huffman coding dictionary
cpdef Integer lzw_compress(str symb):
    cdef bytes symbs = symb.encode()
    cdef dict keys = ascii_to_int.copy()
    cdef int n_keys = 256
    cdef list compressed = []
    cdef int start = 0
    cdef int n_symbs = len(symbs) + 1
    cdef int i
    cdef bytes w
    while True:
        if n_keys >= 512:
            keys = ascii_to_int.copy()
            n_keys = 256
        for i in range(1, n_symbs - start):
            w = symbs[start:start + i]
            if w not in keys:
                compressed.append(keys[w[:-1]])
                keys[w] = n_keys
                start += i - 1
                n_keys += 1
                break
        else:
            compressed.append(keys[w])
            break
    cdef str bits = bit_fill(compressed)
    cdef Integer benc = byte_encode(bits)
    return benc

cpdef str lzw_decompress(Integer encoded):
    cdef bytes encod = encoded.to_bytes(ceil(encoded.bit_length() / 8), "big")
    cdef dict keys = int_to_ascii.copy()
    cdef str bits = byte_decode(encod)
    cdef int n_extended_bytes = floor(len(bits) / 9)
    cdef str bitss = bits[-n_extended_bytes * 9:]
    cdef list data_list = byte2int(bitss, n_extended_bytes)
    cdef bytes previous = keys[data_list[0]]
    cdef list uncompressed = [previous]
    cdef int n_keys = 256
    cdef Integer i
    cdef bytes current
    for i in data_list[1:]:
        if n_keys >= 512:
            keys = int_to_ascii.copy()
            n_keys = 256

        current = keys.get(i)
        if not current:
            current = previous + previous[:1]

        uncompressed.append(current)
        keys[n_keys] = previous + current[:1]
        previous = current
        n_keys += 1

    cdef str unc = b"".join(uncompressed).decode()
    return unc


from math import floor
from .util import *

## Lempel-Ziv-Welch algorithm to compress the symbol sequences in the huffman coding dictionary
def lzw_compress(symbs):
    symbs = symbs.encode()
    keys = ascii_to_int.copy()
    n_keys = 256
    compressed = []
    start = 0
    n_symbs = len(symbs) + 1
    while True:
        if n_keys >= 512:
            keys = ascii_to_int.copy()
            n_keys = 256
        for i in range(1, n_symbs - start):
            w = symbs[start:start + i]
            if w not in keys:
                compressed.append(keys[w[:-1]])
                keys[w] = n_keys
                start += i - 1
                n_keys += 1
                break
        else:
            compressed.append(keys[w])
            break
    bits = bit_fill(compressed)
    return byte_encode(bits, to_bytes = False)

def lzw_decompress(encoded):
    if isinstance(encoded, int):
        encoded = encoded.to_bytes(ceil(encoded.bit_length() / 8), "big")
    keys = int_to_ascii.copy()
    bits = byte_decode(encoded)
    n_extended_bytes = floor(len(bits) / 9)
    bits = bits[-n_extended_bytes * 9:]
    data_list = byte2int(bits, n_extended_bytes)
    previous = keys[data_list[0]]
    uncompressed = [previous]
    n_keys = 256
    for i in data_list[1:]:
        if n_keys >= 512:
            keys = int_to_ascii.copy()
            n_keys = 256

        current = keys.get(i)
        if not current:
            current = previous + previous[:1]

        uncompressed.append(current)
        keys[n_keys] = previous + current[:1]
        previous = current
        n_keys += 1

    return b"".join(uncompressed).decode()



import lzhw
from sys import getsizeof
from random import sample, choices
import pandas as pd

def test_weather():
    weather = ["Sunny", "Sunny", "Overcast", "Rain", "Rain", "Rain", "Overcast", "Sunny", "Sunny",
               "Rain", "Sunny", "Overcast", "Overcast", "Rain", "Rain", "Sunny", "Sunny"]
    comp_weather = lzhw.LZHW(weather)
    assert getsizeof(weather) > comp_weather.size()
    assert weather == comp_weather.decompress()

def test_num():
    numbers = choices(sample(range(0, 5), 5), k = 20)
    comp_num = lzhw.LZHW(numbers)
    assert getsizeof(numbers) > comp_num.size()
    assert numbers == list(map(int, comp_num.decompress()))

def test_read_write():
    weather = ["Sunny", "Sunny", "Overcast", "Rain", "Rain", "Rain", "Overcast", "Sunny", "Sunny",
               "Rain", "Sunny", "Overcast", "Overcast", "Rain", "Rain", "Sunny", "Sunny"]
    comp_weather = lzhw.LZHW(weather)
    comp_weather.save_to_file("test.pkl")
    decomp = lzhw.decompress_from_file("test.pkl")
    assert weather == decomp

def test_comp_df():
    df = pd.DataFrame({"a": [1, 1, 2, 2, 1, 3, 4, 4],
                       "b": ["A", "A", "B", "B", "A", "C,D", "D C", "D C"]})
    comp_df = lzhw.CompressedDF(df)
    assert comp_df.compressed[1].decompress() == list(map(str, df.b))


from collections import Counter
from sys import getsizeof
from pickle import dump, load, HIGHEST_PROTOCOL
#from .lz20 import lz20
from lz20c import lz20
from .huffman_coding import huffman_coding
from .compress_util import code_filling, huffman_decode, org_shaping

# Putting everything together in one class
# Lempel-Ziv-Huffman-Welch. I invented it :D
class LZHW:
    def __init__(self, uncompressed):
        self._compress(uncompressed)

    def _compress(self, uncompressed):
        self.__original_size = getsizeof(uncompressed)
        lz20_sequences = lz20(uncompressed)
        seq_freq = Counter(lz20_sequences)
        huff_coding = huffman_coding(seq_freq)
        self.sequences, self.__codes = code_filling(huff_coding)
        self.compressed = self.__encode(lz20_sequences)

    def __encode(self, sequences):
        # adding preceding 1 to ensure not losing preceding binary 0s
        bitstring = "1" + "".join(self.__codes[seq] for seq in sequences)
        # saving compressed object as an integer from the bit string to save more space
        return int(bitstring, 2)

    def decompress(self):
        decomp = org_shaping(self.sequences, self.compressed)
        return decomp

    def size(self):
        return getsizeof(self.compressed)

    def space_saving(self):
        return "space saving from original to compressed is {}%".format(
            round((1 - (self.size() / self.__original_size)) * 100), 2)

    def save_to_file(self, file):
        with open(file, "wb") as output:
            dump(self.compressed, output, HIGHEST_PROTOCOL)
            dump(self.sequences, output, HIGHEST_PROTOCOL)

    def __repr__(self):
        return bin(self.compressed)[2:]

def decompress_from_file(file):
    with open(file, "rb") as input:
        bit_string = load(input)
        sequences = load(input)
    org =  org_shaping(sequences, bit_string)
    return org



from bisect import insort
from collections import deque

## to store codes for huffman coding
class CodeHeap():
    def __init__(self):
        self._container = deque([])

    def push(self, item):
        insort(self._container, item)  # in by sort

    def pop(self):
        return self._container.popleft()  # FIFO

    def __getitem__(self, item):
        return self._container[item]

    def __len__(self):
        return len(self._container)

    def __repr__(self):
        return repr(self._container)



from .lzhw_alg import LZHW, decompress_from_file
from .lzhw_df import CompressedDF, decompress_df_from_file
from .huffman_coding import huffman_coding
#from .lz20 import lz20
from lz20c import lz20
from .lz78 import lz78
from lzw_c import lzw_compress, lzw_decompress


from .pq import CodeHeap

def huffman_coding(symb2freq):
    """Huffman coding to encode symbols/sequences.
    It takes the Counter of Compressed object and
    returns a code dictionary based on occurences"""

    if len(symb2freq) == 1:
        symb, _ = symb2freq
        return {symb: '0'}

    symb_code_queue = CodeHeap()
    code_dict = {symbs: "" for symbs in symb2freq}

    for symbs, freq in symb2freq.items():
        symb_code_queue.push((freq, [symbs]))

    while len(symb_code_queue) > 1:
        low_freq, low_symbs = symb_code_queue.pop()
        high_freq, high_symbs = symb_code_queue.pop()

        for symb in low_symbs:
            code_dict[symb] = "0" + code_dict[symb]
        for symb in high_symbs:
            code_dict[symb] = "1" + code_dict[symb]

        symb_code_queue.push(
            (low_freq + high_freq,
             sorted(low_symbs + high_symbs)
             )
        )

    return code_dict



from .lzhw_alg import LZHW
from pickle import dump, load, HIGHEST_PROTOCOL
from .compress_util import huffman_decode, org_shaping
import pandas as pd
from lzw_c import *
from tqdm import tqdm
from operator import itemgetter 

class CompressedDF:
    def __init__(self, df, selected_cols = "all"):
        #self.columns = list(df.columns)
        if selected_cols == "all":
            selected = range(df.shape[1])
        else:
            selected = selected_cols
        self.columns = list(itemgetter(*selected)(df.columns))
        self.compressed = []
        for i in tqdm(selected):
            comp_col = LZHW(list(df.iloc[:, i]))
            self.compressed.append(comp_col)

    def save_to_file(self, file):
        with open(file, "wb") as output:
            cols = [i.replace(" ", "__") for i in self.columns]
            dump(lzw_compress(" ".join(cols)), output, HIGHEST_PROTOCOL)
            for i in range(len(self.columns)):
                dump(self.compressed[i].compressed, output, HIGHEST_PROTOCOL)
                dump(self.compressed[i].sequences, output, HIGHEST_PROTOCOL)

def decompress_df_from_file(file, selected_cols = "all"):
    with open(file, "rb") as input:
        cols = load(input)
        cols = lzw_decompress(cols).split()
        cols = [i.replace("__", " ") for i in cols]
        if selected_cols == "all":
            selected = range(len(cols))
        else:
            if any(isinstance(c, str) for c in selected_cols):
                selected_cols = [i for i, v in enumerate(cols) if v in selected_cols]
            selected = selected_cols
        df = {}
        for i in tqdm(range(len(cols))):
            bit_string = load(input)
            sequences = load(input)
            if i in selected:
                df[cols[i]] = org_shaping(sequences, bit_string)
            else:
                continue
    return pd.DataFrame(df)



from lzw_c import *

def glue_seq(seq, last_separate = False):
    if last_separate:
        s = seq.split()
        return " ".join(s[:-1]), s[-1]
    else:
        return " ".join(seq)

def code_filling(huff_codes):
    sequences = {}
    codes = {}
    for seq, code in huff_codes.items():
        cd = int("1" + code, 2)
        sequences[cd] = lzw_compress(seq)
        codes[seq] = code
    return sequences, codes

def huffman_decode(sequences, compressed):
    bitstring = bin(compressed)[2:]
    bit = ""
    for b in bitstring[1:]:  # after preceding 1
        bit += b
        bit_int = int("1" + bit, 2)
        if bit_int in sequences:
            yield lzw_decompress(sequences[bit_int])
            bit = ""

def org_shaping(seq, bits):
    org = " ".join(huffman_decode(seq, bits)).split()
    return [i.replace("__", " ") for i in org]


ctypedef fused str_or_list:
    list
    str

ctypedef fused int_str:
    int
    str

cdef tuple[str] glue_seq_s(str_or_list seq):
    cdef list s = seq.split()
    cdef str h = " ".join(s[:-1])
    cdef str t = s[-1]
    #cdef bytes hb = h.encode()
    #cdef char* hc = hb
    return h, t
		
cdef str glue_seq_l(str_or_list seq):
    cdef str hs = " ".join(seq)
    return hs

cdef list lz20_c(list data):
    data = list(map(str, data))
    cdef str ch
    data = [ch.replace(" ", "__") for ch in data]
    cdef int i = 1
    cdef list sequences = []
    cdef list dst
    cdef int start = 0
    cdef int n_data = len(data)
    cdef dict seq_dict = {}
    cdef str h, t, s
    cdef int ls
    
    while i <= n_data:
        dst = data[start:i]
        s = glue_seq_l(dst)
        ls = len(s.split())
        if s in seq_dict:
            if ls + start >= n_data:
                # if it is the final seq just append it
                sequences.append(s)
                seq_dict[s] += 1
                break  # ensure that it doesn't loop forever
            seq_dict[s] += 1
            i += 1
            continue
        else:
            seq_dict[s] = 1
            if " " in s:  # it is a sequence of symbols not one symbol
                h, t = glue_seq_s(s)
                sequences.append(h)
                sequences.append(t)
            else:
                sequences.append(s)
            start += ls
            i += 1

    return sequences

cpdef list lz20(list uncomp):
    cdef list comp = lz20_c(uncomp)
    return comp



#!/usr/bin/env python

import lzhw
import pandas as pd
import argparse
import csv
import os

parser = argparse.ArgumentParser(description="Data Frame Compressor")
parser.add_argument("-d", "--decompress", help="decompress input into output",
                    action="store_true", default=False)
parser.add_argument("-f", "--input", help="input file to be (de)compressed",
                    type=str, required=True)
parser.add_argument("-o", "--output", help="output where to save result",
                    type=str, required=True)
parser.add_argument("-c", "--columns", nargs="+",
                    help="select specific columns by names or indices (1-based)", type=str,
                    required=False)
parser.add_argument("-nh", "--no-header", help="skip header / data has no header",
                    action="store_true", default=False)
args = vars(parser.parse_args())

file = args["input"]
output = args["output"]

if args["columns"]:
    cols = args["columns"][0]#.split(",")
else:
    cols = "all"

if args["decompress"]:
    if cols != "all":
        cols = [int(i) - 1 for i in cols.split(",")]
    decompressed = lzhw.decompress_df_from_file(file, cols)
    if "xls" in output:
        decompressed.to_excel(output, index=False)
    if "csv" in output:
        decompressed.to_csv(output, index=False)
    else:
        with open(output, "w") as o:
            decompressed.to_string(o, index=False)
    print("decompressed successfully")

else:
    if "xls" in file:
        if args["columns"]:
            command = f"in2csv {file} | csvcut -c {cols}"
            csvdf = csv.reader(os.popen(command).read().splitlines())
            if args["no_header"]:
                h = None
            else:
                h = next(csvdf)
            data = pd.DataFrame(csvdf, columns=h)
        else:
            data = pd.read_excel(file)

    elif "csv" in file:
        if args["columns"]:
            command = f"csvcut -c {cols} {file}"
            csvdf = csv.reader(os.popen(command).read().splitlines())
            if args["no_header"]:
                h = None
            else:
                h = next(csvdf)
            data = pd.DataFrame(csvdf, columns=h)
        else:
            data = pd.read_csv(file)

    else:
        with open(file, "r") as i:
            data = i.read()

    comp_df = lzhw.CompressedDF(data)
    comp_df.save_to_file(output)
    print("compressed successfully")



from setuptools import setup, Extension

cython_exists = False
try:
    from Cython.Distutils import build_ext
    cython_exists = True
except ImportError:
    from distutils.command.build_ext import build_ext

if cython_exists:
    ext_modules=[
        Extension("lz20c", ["lzhw/lz20c.pyx"]),
        Extension("lzw_c", ["lzhw/lzw_c.pyx"])
    ]
else:
    ext_modules = [
        Extension("lz20c", ["lzhw/lz20c.c"]),
        Extension("lzw_c", ["lzhw/lzw_c.c"])
    ]


with open("README.md", "r", encoding="utf8") as rm:
    readme = rm.read()
    
with open("requirements.txt") as rq:
    requirements = rq.read().split('\n')

setup(
      name="lzhw",
      version="1.1.4", 
      description="Compression using an optimized algorithm (lzhw) developed from Lempel-Ziv, Huffman and LZ-Welch techniques",
      packages=["lzhw"],
      install_requires=requirements,
      long_description=readme,
      include_package_data=True,
      long_description_content_type="text/markdown",
      url="https://github.com/MNoorFawi/lzhw",
      author="Muhammad N. Fawi",
      author_email="m.noor.fawi@gmail.com",
      cmdclass = {"build_ext": build_ext},
      license="MIT",
      classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
      ext_modules = ext_modules
)



from math import ceil

def bit_fill(comp):
    return "1" + "".join([bin(i)[2:].zfill(9) for i in comp])

def byte_encode(bits, to_bytes = True):
    if to_bytes:
        return int(bits, 2).to_bytes(ceil(len(bits) / 8), "big")
    else:
        return int(bits, 2)

def byte_decode(enc, from_bytes = True):
    if from_bytes:
        return bin(int.from_bytes(enc, "big"))[2:].zfill(len(enc) * 8)
    else:
        return bin(enc)

def byte2int(bits, ext_bytes):
    return [int(bits[i * 9:(i + 1) * 9], 2) for i in range(ext_bytes)]

ascii_to_int = {i.to_bytes(1, "big"): i for i in range(256)}
int_to_ascii = {i: b for b, i in ascii_to_int.items()}
